/*
 * dump2er.c
 * WiFi Data Elements json objects dump functions
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See AUTHORS file for author and contributors.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>
#include <wifiutils.h>

#include <cmdu.h>

#include "timer.h"
#include "utils.h"

#include "wifi_dataelements.h"
#include "decollector.h"
#include "common.h"
#include "config.h"

/**
 *  Returns: new output lenght
 */
static int remove_new_lines(const char *src, int src_len, char *out, int out_len)
{
	int i, new_len;

	if (out_len < src_len)
		return 0;

	for (i = 0, new_len = 0; i < src_len; ++i) {
		if (src[i] != '\n') {
			*out = src[i];
			++out;
			++new_len;
		}
	}

	return new_len;
}

static const char *to_base64_str(const char *src, size_t src_len, char *out,
				 size_t out_len)
{
	size_t tmp_len = out_len;
	char *tmp_str;
	int new_out_len;

	if (!out || !src)
		return NULL;

	tmp_str = malloc(tmp_len);
	if (!tmp_str)
		return NULL;

	if (base64_encode((const unsigned char *)src, src_len,
			  (unsigned char *)tmp_str, &tmp_len)) {
		fprintf(stderr, "%s: base64_encode failed.", __func__);
		free(tmp_str);
		return NULL;
	}

	new_out_len = remove_new_lines(tmp_str, tmp_len, out, out_len);
	free(tmp_str);

	if (new_out_len < 1)
		return NULL;

	if (new_out_len < out_len)
		out[new_out_len] = '\0';
	else
		out[new_out_len - 1] = '\0';

	return out;
}

static char *create_band_list(char *str_buf, size_t str_len, uint32_t band)
{
	char *str = str_buf;
	ssize_t len = 0;

#define ADD_BAND(b) do { \
	len = snprintf(str, str_len, "%s%s", len > 0 ? "," : "", b); \
	if (len < 0 || len >= str_len) { \
		fprintf(stderr, "snprintf fail to add band"); \
		return str_buf; \
	} \
	str += len; \
	str_len -= len; \
} while (0)

	if (band & BAND_2)
		ADD_BAND("2");
	if (band & BAND_5)
		ADD_BAND("5");
	if (band & BAND_6)
		ADD_BAND("6");
	if (band & BAND_60)
		ADD_BAND("60");

#undef ADD_BAND

	return str_buf;
}

static void dump2_network_device_default8021q(struct decollector_private *p,
					      struct wifi_network *net,
					      struct wifi_network_device *dev,
					      struct blob_buf *bb)
{
	void *arr, *t;
	int i;

	/* TODO this is for device or for network ? */

	/* Device.WiFi.DataElements.Network.Device.{i}.Default8021Q.{i}. */
	arr = blobmsg_open_array(bb, "Default8021Q");
	for (i = 0; i < 1; i++) {
		t = blobmsg_open_table(bb, "");
		blobmsg_add_u8(bb, "Enable", net->primary_vid ? true : false);
		blobmsg_add_u32(bb, "PrimaryVID", net->primary_vid);
		blobmsg_add_u32(bb, "DefaultPCP", net->default_pcp);
		blobmsg_close_table(bb, t);
	}

	blobmsg_close_array(bb, arr); // Default8021Q
}


static void dump2_network_device_ssidtovidmapping(struct decollector_private *p,
					      struct wifi_network *net,
					      struct wifi_network_device *dev,
					      struct blob_buf *bb)
{
	struct wifi_network_ssid *ssid;
	void *arr, *t;

	/* TODO this is for device or for network ? */

	/* Device.WiFi.DataElements.Network.Device.{i}.SSIDtoVIDMapping.{i}. */
	arr = blobmsg_open_array(bb, "SSIDtoVIDMapping");
	list_for_each_entry(ssid, &net->ssidlist, list) {
		t = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "SSID", (char *)ssid->ssid);
		blobmsg_add_u32(bb, "VID", ssid->vid);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr); // SSIDtoVIDMapping

}


static void dump2_network_device_cacstatus(struct decollector_private *p,
					   struct wifi_network *net,
					   struct wifi_network_device *dev,
					   struct blob_buf *bb)
{
	const struct wifi_cac_status *cac_status_entry;
	void *cac_status_array;

	/* Device.WiFi.DataElements.Network.Device.{i}.CACStatus.{i}. */
	cac_status_array = blobmsg_open_array(bb, "CACStatus");
	list_for_each_entry(cac_status_entry, &dev->cac_statuslist, list) {

		void *cac_status_table = blobmsg_open_table(bb, "");
		void *tmp_array;
		const struct wifi_cac_available_channel *cac_avail_entry;
		const struct wifi_cac_nop_channel *cac_nop_entry;
		const struct wifi_cac_active_channel *cac_active_entry;

		blobmsg_add_string(bb, "TimeStamp", cac_status_entry->tsp);

		blobmsg_add_u32(bb, "CACAvailableChannelNumberOfEntries",
				list_count(&cac_status_entry->available_chlist));
		blobmsg_add_u32(bb, "CACNonOccupancyChannelNumberOfEntries",
				list_count(&cac_status_entry->nop_chlist));
		blobmsg_add_u32(bb, "CACActiveChannelNumberOfEntries",
				list_count(&cac_status_entry->cac_chlist));

		/* Device.WiFi.DataElements.Network.Device.{i}.CACStatus.{i}.CACAvailableChannel.{i}. */
		tmp_array = blobmsg_open_array(bb, "CACAvailableChannel");
		list_for_each_entry(cac_avail_entry, &cac_status_entry->available_chlist, list) {
			void *table = blobmsg_open_table(bb, "");

			blobmsg_add_u32(bb, "OpClass", cac_avail_entry->opclass);
			blobmsg_add_u32(bb, "Channel", cac_avail_entry->channel);
			blobmsg_add_u32(bb, "Minutes", cac_avail_entry->cleared);

			blobmsg_close_table(bb, table);
		}
		blobmsg_close_array(bb, tmp_array); // CACAvailableChannel

		/* Device.WiFi.DataElements.Network.Device.{i}.CACStatus.{i}.CACNonOccupancyChannel.{i}. */
		tmp_array = blobmsg_open_array(bb, "CACNonOccupancyChannel");
		list_for_each_entry(cac_nop_entry, &cac_status_entry->nop_chlist, list) {
			void *table = blobmsg_open_table(bb, "");

			blobmsg_add_u32(bb, "OpClass", cac_nop_entry->opclass);
			blobmsg_add_u32(bb, "Channel", cac_nop_entry->channel);
			blobmsg_add_u32(bb, "Seconds", cac_nop_entry->remaining);

			blobmsg_close_table(bb, table);

		}
		blobmsg_close_array(bb, tmp_array); // CACNonOccupancyChannel

		/* Device.WiFi.DataElements.Network.Device.{i}.CACStatus.{i}.CACActiveChannel.{i}. */
		tmp_array = blobmsg_open_array(bb, "CACActiveChannel");
		list_for_each_entry(cac_active_entry, &cac_status_entry->cac_chlist, list) {
			void *table = blobmsg_open_table(bb, "");

			blobmsg_add_u32(bb, "OpClass", cac_active_entry->opclass);
			blobmsg_add_u32(bb, "Channel", cac_active_entry->channel);
			blobmsg_add_u32(bb, "Countdown", cac_active_entry->remaining);

			blobmsg_close_table(bb, table);

		}
		blobmsg_close_array(bb, tmp_array); // CACActiveChannel

		blobmsg_close_table(bb, cac_status_table); //CACStatus.{i}
	}

	blobmsg_close_array(bb, cac_status_array); // CACStatus

}

static void dump2_network_device_sprule(struct decollector_private *p,
					struct wifi_network *net,
					struct wifi_network_device *dev,
					struct blob_buf *bb)
{
	void *arr, *t;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.SPRule.{i}. */
	arr = blobmsg_open_array(bb, "SPRule");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");

		blobmsg_add_u32(bb, "ID", 0); // TODO unsignedInt
		blobmsg_add_u32(bb, "Precedence", 0); // TODO unsignedInt​(:254)
		blobmsg_add_u32(bb, "Output", 0); // TODO unsignedInt​(:9)
		blobmsg_add_string(bb, "AlwaysMatch", "TODO bool"); // TODO boolean

		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);
}

static void dump2_network_device_ieee1905security(struct decollector_private *p,
						  struct wifi_network *net,
						  struct wifi_network_device *dev,
						  struct blob_buf *bb)
{
	void *arr;
	void *t;
	int i;
	const int num_of_entries = dev->i1905_seccap.caps_valid ? 1 : 0;

	/* Device.WiFi.DataElements.Network.Device.{i}.IEEE1905Security.{i}. */
	arr = blobmsg_open_array(bb, "IEEE1905Security");
	for (i = 0; i < num_of_entries; i++) {
		t = blobmsg_open_table(bb, "");

		blobmsg_add_u32(bb, "OnboardingProtocol", dev->i1905_seccap.onboarding_protocol);
		blobmsg_add_u32(bb, "IntegrityAlgorithm", dev->i1905_seccap.integrity);
		blobmsg_add_u32(bb, "EncryptionAlgorithm", dev->i1905_seccap.encryption);

		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr); // IEEE1905Security

}

static void dump2_network_device_anticipatedchannels(struct decollector_private *p,
						     struct wifi_network *net,
						     struct wifi_network_device *dev,
						     struct blob_buf *bb)
{
	void *arr, *t;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.AnticipatedChannels.{i}. */
	arr = blobmsg_open_array(bb, "AnticipatedChannels");

	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");

		blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)

		arr = blobmsg_open_array(bb, "ChannelList");
		for (i = 0; i < 0 /* TODO */; i++) {
			 ; // list of unsignedInt​(:255)
		}
		blobmsg_close_array(bb, arr);

		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr); // AnticipatedChannels
}

static void dump2_network_device_anticipatedchannelsusage(struct decollector_private *p,
						     struct wifi_network *net,
						     struct wifi_network_device *dev,
						     struct blob_buf *bb)
{
	void *arr, *usage_table, *entry_table;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.AnticipatedChannelUsage.{i}. */
	arr = blobmsg_open_array(bb, "AnticipatedChannelUsage");
	for (i = 0; i < 0 /* TODO */; i++) {

		usage_table = blobmsg_open_table(bb, "");

		blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
		//hwaddr_ntoa(usta->macaddr, macaddr);)
		blobmsg_add_string(bb, "ReferenceBSSID", "TODO MACAddress");
		blobmsg_add_u32(bb, "EntryNumberOfEntries", 0); // TODO unsignedInt

		/* Device.WiFi.DataElements.Network.Device.{i}.AnticipatedChannelUsage.{i}.Entry.{i}. */
		arr = blobmsg_open_array(bb, "Entry");
		for (i = 0; i < 0 /* TODO */; i++) {
			entry_table = blobmsg_open_table(bb, "");

			blobmsg_add_string(bb, "BurstStartTime", "TODO hexBinary​(4:4)");
			blobmsg_add_u32(bb, "BurstLength", 0); // TODO unsignedInt
			blobmsg_add_u32(bb, "Repetitions", 0); // TODO unsignedInt
			blobmsg_add_u32(bb, "BurstInterval", 0); // TODO unsignedInt
			blobmsg_add_string(bb, "RUBitmask", "TODO hexBinary​(2:10)");
			//hwaddr_ntoa(usta->macaddr, macaddr);)
			blobmsg_add_string(bb, "TransmitterIdentifier", "TODO MACAddress");
			blobmsg_add_u32(bb, "PowerLevel", 0); // TODO int
			blobmsg_add_string(bb, "ChannelUsageReason", "TODO");

			blobmsg_close_table(bb, entry_table);
		}
		blobmsg_close_array(bb, arr); // Entry

		blobmsg_close_table(bb, usage_table);
	}

	blobmsg_close_array(bb, arr); // AnticipatedChannelUsage
}


static void dump2_network_device_multiapdevice(struct decollector_private *p,
					       struct wifi_network *net,
					       struct wifi_network_device *dev,
					       struct blob_buf *bb)
{
	void *map_device, *backhaul, *class_profile, *stats;
	void *class_profile_arr;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.MultiAPDevice. */
	map_device = blobmsg_open_table(bb, "MultiAPDevice");
	blobmsg_add_string(bb, "ManufacturerOUI", "TODO");
	blobmsg_add_string(bb, "LastContactTime", "TODO dateTime");
	blobmsg_add_string(bb, "AssocIEEE1905DeviceRef", "TODO");
	blobmsg_add_string(bb, "EasyMeshControllerOperationMode", "TODO");
	blobmsg_add_string(bb, "EasyMeshAgentOperationMode", "TODO");


	/* Device.WiFi.DataElements.Network.Device.{i}.MultiAPDevice.Backhaul. */
	backhaul = blobmsg_open_table(bb, "Backhaul");
	blobmsg_add_string(bb, "LinkType", "TODO");  //TODO string​
	//hwaddr_ntoa(usta->macaddr, macaddr);)
	blobmsg_add_string(bb, "BackhaulMACAddress", "TODO MACAddress");  //TODO string​(17)
	//hwaddr_ntoa(usta->macaddr, macaddr);)
	blobmsg_add_string(bb, "BackhaulDeviceID", "TODO MACAddress");  //TODO string​(17)
	//hwaddr_ntoa(usta->macaddr, macaddr);)
	blobmsg_add_string(bb, "MACAddress", "TODO MACAddress"); //TODO string​(17)
	blobmsg_add_u32(bb, "CurrentOperatingClassProfileNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.MultiAPDevice.Backhaul.CurrentOperatingClassProfile.{i}. */
	class_profile_arr = blobmsg_open_array(bb, "CurrentOperatingClassProfileList");
	for (i = 0; i < 0 /* TODO */; ++i) {
		class_profile = blobmsg_open_table(bb, "CurrentOperatingClassProfile");

		blobmsg_add_u32(bb, "Class", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "TxPower", 0); // TODO int​(-127:127)
		blobmsg_add_string(bb, "TimeStamp", "TODO dateTime"); //TODO "2002-05-30T09:00:00"
		blobmsg_close_table(bb, class_profile);
	}
	blobmsg_close_array(bb, class_profile_arr);

	/* Device.WiFi.DataElements.Network.Device.{i}.MultiAPDevice.Backhaul.Stats. */
	stats = blobmsg_open_table(bb, "Stats");
	blobmsg_add_u64(bb, "BytesSent", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BytesReceived", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "PacketsSent", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "PacketsReceived", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "ErrorsSent", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "ErrorsReceived", 0); //TODO StatsCounter64
	blobmsg_add_u32(bb, "LinkUtilization", 0); // TODO unsignedInt​(:100)
	blobmsg_add_u32(bb, "SignalStrength", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "LastDataDownlinkRate", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "LastDataUplinkRate", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "TimeStamp", "TODO dateTime"); //TODO "2002-05-30T09:00:00"
	blobmsg_close_table(bb, stats);

	blobmsg_close_table(bb, backhaul);
	blobmsg_close_table(bb, map_device);

}

static void dump2_network_device_radio_scanresult(struct decollector_private *p,
						  struct wifi_network *net,
						  struct wifi_network_device *dev,
						  struct wifi_radio_element *radio,
						  struct blob_buf *bb)
{
	struct wifi_scanres_element *sres = NULL;
	void *arr, *arr2, *arr3, *arr4;
	void *t, *t2, *t3, *t4;


	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.ScanResult.{i}. */
	arr = blobmsg_open_array(bb, "ScanResult");
	list_for_each_entry(sres, &radio->scanlist, list) {
		struct wifi_scanres_opclass_element *op = NULL;

		t = blobmsg_open_table(bb, "");
		blobmsg_add_string(bb, "TimeStamp", sres->tsp);
		blobmsg_add_u32(bb, "OpClassScanNumberOfEntries", sres->num_opclass_scanned);

		/* ... Radio.{i}.ScanResult.{i}.OpClassScan.{i}. */
		arr2 = blobmsg_open_array(bb, "OpClassScan");
		list_for_each_entry_reverse(op, &sres->opclass_scanlist, list) {
			struct wifi_scanres_channel_element *ch = NULL;

			t2 = blobmsg_open_table(bb, "");
			blobmsg_add_u32(bb, "OperatingClass", op->opclass);
			blobmsg_add_u32(bb, "ChannelScanNumberOfEntries", op->num_channels_scanned);

			/* ... ScanResult.{i}.OpClassScan.{i}.ChannelScan.{i}. */
			arr3 = blobmsg_open_array(bb, "ChannelScan");
			list_for_each_entry(ch, &op->channel_scanlist, list) {
				struct wifi_scanres_neighbor_element *nbr = NULL;

				t3 = blobmsg_open_table(bb, "");
				blobmsg_add_u32(bb, "Channel", ch->channel);
				blobmsg_add_string(bb, "TimeStamp", ch->tsp);
				blobmsg_add_u32(bb, "Utilization", ch->utilization);
				blobmsg_add_u32(bb, "Noise", ch->anpi);
				blobmsg_add_u32(bb, "NeighborBSSNumberOfEntries", ch->num_neighbors);

				/* ... OpClassScan.{i}.ChannelScan.{i}.NeighborBSS.{i}. */
				arr4 = blobmsg_open_array(bb, "NeighborBSS");
				list_for_each_entry(nbr, &ch->nbrlist, list) {
					char nbr_bssid[18] = { 0 };

					t4 = blobmsg_open_table(bb, "");
					hwaddr_ntoa(nbr->bssid, nbr_bssid);
					blobmsg_add_string(bb, "BSSID", nbr_bssid);
					blobmsg_add_string(bb, "SSID", nbr->ssid);
					blobmsg_add_u32(bb, "SignalStrength", nbr->rssi);
					blobmsg_add_u32(bb, "ChannelBandwidth", nbr->bw);
					blobmsg_add_u32(bb, "ChannelUtilization", nbr->utilization);
					blobmsg_add_u32(bb, "StationCount", nbr->num_stations);
					blobmsg_close_table(bb, t4);
				}
				blobmsg_close_array(bb, arr4); // NeighborBSS
				blobmsg_close_table(bb, t3);
			}
			blobmsg_close_array(bb, arr3); // ChannelScan
			blobmsg_close_table(bb, t2);
		}
		blobmsg_close_array(bb, arr2); // OpClassScan
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr); // ScanResult
}

static void dump2_network_device_radio_bss_sta_multiapsta(struct decollector_private *p,
							  struct wifi_network *net,
							  struct wifi_network_device *dev,
							  struct wifi_radio_element *radio,
							  struct wifi_bss_element *bss,
							  struct wifi_sta_element *sta,
							  struct blob_buf *bb)
{
	void *t, *t2;
	void *arr;
	int i;


	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.MultiAPSTA. */
	t = blobmsg_open_table(bb, "MultiAPSTA");
	blobmsg_add_string(bb, "AssociationTime", "TODO dateTime");
	blobmsg_add_u32(bb, "Noise", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "SteeringHistoryNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.MultiAPSTA.SteeringSummaryStats. */
	t2 = blobmsg_open_table(bb, "SteeringSummaryStats");
	blobmsg_add_u64(bb, "NoCandidateAPFailures", sta->mapsta.stats.no_candidate_cnt);
	blobmsg_add_u64(bb, "BlacklistAttempts", sta->mapsta.stats.blacklist_attempt_cnt);
	blobmsg_add_u64(bb, "BlacklistSuccesses", sta->mapsta.stats.blacklist_success_cnt);
	blobmsg_add_u64(bb, "BlacklistFailures", sta->mapsta.stats.blacklist_failure_cnt);
	blobmsg_add_u64(bb, "BTMAttempts", sta->mapsta.stats.btm_attempt_cnt);
	blobmsg_add_u64(bb, "BTMSuccesses", sta->mapsta.stats.btm_success_cnt);
	blobmsg_add_u64(bb, "BTMFailures", sta->mapsta.stats.btm_failure_cnt);
	blobmsg_add_u64(bb, "BTMQueryResponses", sta->mapsta.stats.btm_query_resp_cnt);
	blobmsg_add_u32(bb, "LastSteerTime", sta->mapsta.stats.last_tsp.tv_sec);
	blobmsg_close_table(bb, t2); // SteeringSummaryStats

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.MultiAPSTA.SteeringHistory.{i}. */
	arr = blobmsg_open_array(bb, "SteeringHistory");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "Time", "TODO dateTime");
		blobmsg_add_string(bb, "APOrigin", "TODO");
		blobmsg_add_string(bb, "TriggerEvent", "TODO");
		blobmsg_add_string(bb, "SteeringApproach", "TODO");
		blobmsg_add_string(bb, "APDestination", "TODO");
		blobmsg_add_u32(bb, "SteeringDuration", 0); // TODO unsignedInt
	}
	blobmsg_close_array(bb, arr); // SteeringHistory

	blobmsg_close_table(bb, t); // MultiAPSTA
}

static void dump2_network_device_radio_bss_sta_wifi6cap(struct decollector_private *p,
							struct wifi_network *net,
							struct wifi_network_device *dev,
							struct wifi_radio_element *radio,
							struct wifi_bss_element *bss,
							struct wifi_sta_element *sta,
							struct blob_buf *bb)
{
	void *t;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.WiFi6Capabilities. */
	t = blobmsg_open_table(bb, "WiFi6Capabilities");
	blobmsg_add_string(bb, "HE160", "TODO bool");
	blobmsg_add_string(bb, "HE8080", "TODO bool");
	blobmsg_add_string(bb, "MCSNSS", "TODO base64​(4:12)");
	blobmsg_add_string(bb, "SUBeamformer", "TODO bool");
	blobmsg_add_string(bb, "SUBeamformee", "TODO bool");
	blobmsg_add_string(bb, "MUBeamformer", "TODO bool");
	blobmsg_add_string(bb, "Beamformee80orLess", "TODO bool");
	blobmsg_add_string(bb, "BeamformeeAbove80", "TODO bool");
	blobmsg_add_string(bb, "ULMUMIMO", "TODO bool");
	blobmsg_add_string(bb, "ULOFDMA", "TODO bool");
	blobmsg_add_u32(bb, "MaxDLMUMIMO", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxULMUMIMO", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxDLOFDMA", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "MaxULOFDMA", 0); // TODO unsignedInt​(:255)
	blobmsg_add_string(bb, "RTS", "TODO bool");
	blobmsg_add_string(bb, "MURTS", "TODO bool");
	blobmsg_add_string(bb, "MultiBSSID", "TODO bool");
	blobmsg_add_string(bb, "MUEDCA", "TODO bool");
	blobmsg_add_string(bb, "TWTRequestor", "TODO bool");
	blobmsg_add_string(bb, "TWTResponder", "TODO bool");
	blobmsg_add_string(bb, "SpatialReuse", "TODO bool");
	blobmsg_add_string(bb, "AnticipatedChannelUsage", "TODO bool");
	blobmsg_close_table(bb, t); // WiFi6Capabilities
}

static void dump2_network_device_radio_bss_sta(struct decollector_private *p,
					       struct wifi_network *net,
					       struct wifi_network_device *dev,
					       struct wifi_radio_element *radio,
					       struct wifi_bss_element *bss,
					       struct wifi_sta_element *sta,
					       struct blob_buf *bb)
{
	char ip_addr[INET6_ADDRSTRLEN] = { 0 };
	char caps_b64[32] = { 0 };
	char macaddr[18] = { 0 };
	const char *caps_b64_ret = NULL;
	int he_caps_len;
	void *arr;
	void *t;
	int i;


	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}. */
	hwaddr_ntoa(sta->macaddr, macaddr);
	blobmsg_add_string(bb, "MACAddress", macaddr);
	blobmsg_add_string(bb, "TimeStamp", sta->tsp);

	caps_b64_ret = to_base64_str((const char *)&sta->caps.ht,
				     sizeof(sta->caps.ht), caps_b64,
				     sizeof(caps_b64));
	blobmsg_add_string(bb, "HTCapabilities", caps_b64_ret ? caps_b64_ret : "");

	caps_b64_ret = to_base64_str((const char *)&sta->caps.vht,
				     sizeof(sta->caps.vht), caps_b64,
				     sizeof(caps_b64));
	blobmsg_add_string(bb, "VHTCapabilities", caps_b64_ret ? caps_b64_ret : "");

	if (sta->caps.valid & HE_CAP_VALID)
		he_caps_len = 2 + sta->caps.he[0];
	else
		he_caps_len = 4; /* pass the usp-validator range test */

	caps_b64_ret = to_base64_str((const char *)&sta->caps.he[1],
				     he_caps_len, caps_b64,
				     sizeof(caps_b64));
	blobmsg_add_string(bb, "HECapabilities", caps_b64_ret ? caps_b64_ret : "");

	blobmsg_add_string(bb, "ClientCapabilities", "TODO base64");
	blobmsg_add_u32(bb, "LastDataDownlinkRate", sta->dl_rate);
	blobmsg_add_u32(bb, "LastDataUplinkRate", sta->ul_rate);
	blobmsg_add_u32(bb, "UtilizationReceive", sta->dl_utilization);
	blobmsg_add_u32(bb, "UtilizationTransmit", sta->ul_utilization);
	blobmsg_add_u32(bb, "EstMACDataRateDownlink", sta->dl_est_thput);
	blobmsg_add_u32(bb, "EstMACDataRateUplink", sta->ul_est_thput);
	blobmsg_add_u32(bb, "SignalStrength", sta->rcpi);
	blobmsg_add_u32(bb, "LastConnectTime", sta->conn_time);
	blobmsg_add_u64(bb, "BytesSent", sta->tx_bytes);
	blobmsg_add_u64(bb, "BytesReceived", sta->rx_bytes);
	blobmsg_add_u64(bb, "PacketsSent", sta->tx_pkts);
	blobmsg_add_u64(bb, "PacketsReceived", sta->rx_pkts);
	blobmsg_add_u64(bb, "ErrorsSent", sta->tx_errors);
	blobmsg_add_u64(bb, "ErrorsReceived", sta->rx_errors);
	blobmsg_add_u64(bb, "RetransCount", sta->rtx_pkts);

	blobmsg_add_u32(bb, "NumberOfMeasureReports", sta->num_meas_reports);

	arr = blobmsg_open_array(bb, "MeasurementReport");
	for (i = 0; i < sta->num_meas_reports; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of base64
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	blobmsg_add_string(bb, "IPV4Address",
			   inet_ntop(AF_INET, &sta->ipv4_addr.addr.ip4, ip_addr, INET_ADDRSTRLEN));
	blobmsg_add_string(bb, "IPV6Address",
			   inet_ntop(AF_INET6, &sta->ipv6_addr.addr.ip6, ip_addr, INET6_ADDRSTRLEN));

	blobmsg_add_string(bb, "Hostname", sta->hostname);
	blobmsg_add_string(bb, "CellularDataPreference", "TODO");
	blobmsg_add_u32(bb, "ReAssociationDelay", 0); // TODO unsignedInt​(:65535)
	blobmsg_add_u32(bb, "TIDQueueSizesNumberOfEntries", 0); // TODO unsignedInt

	dump2_network_device_radio_bss_sta_multiapsta(p, net, dev, radio, bss, sta, bb);
	dump2_network_device_radio_bss_sta_wifi6cap(p, net, dev, radio, bss, sta, bb);

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.STA.{i}.TIDQueueSizes.{i}. */
	arr = blobmsg_open_array(bb, "TIDQueueSizes");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "TID", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "Size", 0); // TODO unsignedInt​(:255)
	}
	blobmsg_close_array(bb, arr); // TIDQueueSizes
}


static void dump2_network_device_radio_bss(struct decollector_private *p,
					   struct wifi_network *net,
					   struct wifi_network_device *dev,
					   struct wifi_radio_element *radio,
					   struct wifi_bss_element *bss,
					   struct blob_buf *bb)
{
	struct wifi_sta_element *sta;
	char bssid[18] = { 0 };
	void *arr;
	void *t, *bss_table;
	int i;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}. */

	bss_table = blobmsg_open_table(bb, "");

	hwaddr_ntoa(bss->bssid, bssid);
	blobmsg_add_string(bb, "BSSID", bssid);
	blobmsg_add_string(bb, "SSID", (char *)bss->ssid);
	blobmsg_add_string(bb, "Enabled", bss->enabled ? "true" : "false");
	blobmsg_add_u32(bb, "LastChange", bss->uptime);
	blobmsg_add_string(bb, "TimeStamp", bss->tsp);
	blobmsg_add_u64(bb, "UnicastBytesSent", bss->tx_ucast_bytes);
	blobmsg_add_u64(bb, "UnicastBytesReceived", bss->rx_ucast_bytes);
	blobmsg_add_u64(bb, "MulticastBytesSent", bss->tx_mcast_bytes);
	blobmsg_add_u64(bb, "MulticastBytesReceived", bss->rx_mcast_bytes);
	blobmsg_add_u64(bb, "BroadcastBytesSent", bss->tx_bcast_bytes);
	blobmsg_add_u64(bb, "BroadcastBytesReceived", bss->rx_bcast_bytes);
	blobmsg_add_u32(bb, "ByteCounterUnits", 0); //TODO
	blobmsg_add_string(bb, "Profile1bSTAsDisallowed", "TODO bool");
	blobmsg_add_string(bb, "Profile2bSTAsDisallowed", "TODO bool");
	blobmsg_add_u32(bb, "AssociationAllowanceStatus", 0); // TODO unsignedInt​(0:1)
	blobmsg_add_string(bb, "EstServiceParametersBE", "TODO base64​(0:3)");
	blobmsg_add_string(bb, "EstServiceParametersBK", "TODO base64​(0:3)");
	blobmsg_add_string(bb, "EstServiceParametersVI", "TODO base64​(0:3)");
	blobmsg_add_string(bb, "EstServiceParametersVO", "TODO base64​(0:3)");
	blobmsg_add_u8(bb, "BackhaulUse", bss->is_bbss);
	blobmsg_add_u8(bb, "FronthaulUse", bss->is_fbss);
	blobmsg_add_u8(bb, "R1disallowed", bss->r1_disallowed);
	blobmsg_add_u8(bb, "R2disallowed", bss->r2_disallowed);
	blobmsg_add_u8(bb, "MultiBSSID", bss->multi_bssid);
	blobmsg_add_u8(bb, "TransmittedBSSID", bss->transmitted_bssid);

	arr = blobmsg_open_array(bb, "FronthaulAKMsAllowed");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of string
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	arr = blobmsg_open_array(bb, "BackhaulAKMsAllowed");
	for (i = 0; i < 0 /* TODO */; i++) {
		t = blobmsg_open_table(bb, "");
		 ; // list of string
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	blobmsg_add_u32(bb, "STANumberOfEntries", bss->num_stations);
	blobmsg_add_u32(bb, "QMDescriptorNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.QMDescriptor.{i}. */
	arr = blobmsg_open_array(bb, "QMDescriptor");
	for (i = 0; i < 0 /* TODO */; i++) {
		//hwaddr_ntoa(usta->macaddr, macaddr);)
		blobmsg_add_string(bb, "ClientMAC", "TODO MACAddress");
		blobmsg_add_string(bb, "DescriptorElement", "TODO hexBinary");
	}
	blobmsg_close_array(bb, arr); // QMDescriptor

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BSS.{i}.MultiAPSteering. */
	t = blobmsg_open_table(bb, "MultiAPSteering");
	blobmsg_add_u64(bb, "BlacklistAttempts", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMAttempts", 0); //TODO StatsCounter64
	blobmsg_add_u64(bb, "BTMQueryResponses", 0); //TODO StatsCounter64
	blobmsg_close_table(bb, t); // MultiAPSteering

	arr = blobmsg_open_array(bb, "STAList");
	list_for_each_entry(sta, &bss->stalist, list) {
		t = blobmsg_open_table(bb, "");
		dump2_network_device_radio_bss_sta(p, net, dev, radio, bss, sta, bb);  /* cppcheck-suppress uninitvar */
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, arr);

	blobmsg_close_table(bb, bss_table);
}

static void dump2_network_device_radio_backhaulsta(struct decollector_private *p,
						   struct wifi_network *net,
						   struct wifi_network_device *dev,
						   struct wifi_radio_element *radio,
						   struct blob_buf *bb)
{
	struct wifi_backhaul_element *bsta = &radio->bsta;
	char macaddr[18] = { 0 };
	void *t;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.BackhaulSta. */
	t = blobmsg_open_table(bb, "BackhaulSta");
	hwaddr_ntoa(bsta->macaddr, macaddr);
	blobmsg_add_string(bb, "MACAddress", macaddr);
	blobmsg_close_table(bb, t); // BackhaulSta
}

static void dump2_network_device_radio_scancapability(struct decollector_private *p,
						      struct wifi_network *net,
						      struct wifi_network_device *dev,
						      struct wifi_radio_element *radio,
						      struct blob_buf *bb)
{
	void *scan_cap, *op_class;
	void *op_class_array, *channel_array;
	int i, j;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.ScanCapability. */
	scan_cap = blobmsg_open_table(bb, "ScanCapability");

	blobmsg_add_u8(bb, "OnBootOnly", radio->scan_caps.boot_only);
	blobmsg_add_u32(bb, "Impact", radio->scan_caps.impact);
	blobmsg_add_u32(bb, "MinimumInterval", radio->scan_caps.interval);
	blobmsg_add_u32(bb, "OpClassChannelsNumberOfEntries", radio->scan_caps.opclass.num_opclass);

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.ScanCapability.OpClassChannels.{i}. */
	op_class_array = blobmsg_open_array(bb, "OpClassChannels");

	for (i = 0; i < radio->scan_caps.opclass.num_opclass; i++) {
		op_class = blobmsg_open_table(bb, "");
		blobmsg_add_u32(bb, "OpClass", radio->scan_caps.opclass.opclass[i].id);

		channel_array = blobmsg_open_array(bb, "ChannelList");

		for (j = 0; j < radio->scan_caps.opclass.opclass[i].num_channel; j++)
			blobmsg_add_u32(bb, NULL, radio->scan_caps.opclass.opclass[i].channel[j].channel);

		blobmsg_close_array(bb, channel_array);

		blobmsg_close_table(bb, op_class);
	}
	blobmsg_close_array(bb, op_class_array); // OpClassChannels

	blobmsg_close_table(bb, scan_cap); // ScanCapability
}

static void dump2_network_device_radio_caccapability(struct decollector_private *p,
						     struct wifi_network *net,
						     struct wifi_network_device *dev,
						     struct wifi_radio_element *radio,
						     struct blob_buf *bb)
{
	void *cac_method_arr, *op_class_channels_arr, *channel_arr;
	void *cac_capability, *cac_method, *opclass_channel;
	int i, j;
	struct wifi_radio_cac_capabilities *cac_caps;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.CACCapability. */
	cac_capability = blobmsg_open_table(bb, "CACCapability");
	blobmsg_add_u32(bb, "CACMethodNumberOfEntries", list_count(&radio->cac_capslist));

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.CACCapability.CACMethod.{i}. */
	cac_method_arr = blobmsg_open_array(bb, "CACMethod");
	list_for_each_entry(cac_caps, &radio->cac_capslist, list) {
		cac_method = blobmsg_open_table(bb, "");

		blobmsg_add_u32(bb, "Method", cac_caps->method);
		blobmsg_add_u32(bb, "NumberOfSeconds", cac_caps->num_seconds);
		blobmsg_add_u32(bb, "OpClassChannelsNumberOfEntries", cac_caps->opclasses.num_opclass);

		/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.CACCapability.CACMethod.{i}.OpClassChannels.{i}. */
		op_class_channels_arr = blobmsg_open_array(bb, "OpClassChannels");
		for (i = 0; i < cac_caps->opclasses.num_opclass; ++i) {
			opclass_channel = blobmsg_open_table(bb, "");

			blobmsg_add_u32(bb, "OpClass", cac_caps->opclasses.opclass[i].id);

			channel_arr = blobmsg_open_array(bb, "ChannelList");
			for (j = 0; j < cac_caps->opclasses.opclass[i].num_channel; ++j)
				blobmsg_add_u32(bb, "", cac_caps->opclasses.opclass[i].channel[j].channel);

			blobmsg_close_array(bb, channel_arr);

			blobmsg_close_table(bb, opclass_channel);
		}
		blobmsg_close_array(bb, op_class_channels_arr);

		blobmsg_close_table(bb, cac_method);
	}
	blobmsg_close_array(bb, cac_method_arr);

	blobmsg_close_table(bb, cac_capability);
}

static void dump2_network_device_radio_capabilities_wifi6_role(
	const struct wifi_wifi6_capabilities *role_caps, struct blob_buf *bb)
{
	char caps_b64[32] = {0};
	const char *caps_b64_ret = NULL;

	blobmsg_add_u8(bb, "HE160",  role_caps->he160);
	blobmsg_add_u8(bb, "HE8080", role_caps->he8080);

	/* Encode 4, 8 or 12 bytes of MCS NSS */
	caps_b64_ret = to_base64_str((const char *)&role_caps->mcs_nss_12,
				     role_caps->mcs_nss_len, caps_b64,
				     sizeof(caps_b64));
	blobmsg_add_string(bb, "MCSNSS",  caps_b64_ret ? caps_b64_ret : "");

	blobmsg_add_u8(bb, "SUBeamformer", role_caps->su_beamformer);
	blobmsg_add_u8(bb, "SUBeamformee", role_caps->su_beamformee);
	blobmsg_add_u8(bb, "MUBeamformer", role_caps->mu_beamformer);
	blobmsg_add_u8(bb, "Beamformee80orLess", role_caps->beamformee_le80);
	blobmsg_add_u8(bb, "BeamformeeAbove80", role_caps->beamformee_gt80);
	blobmsg_add_u8(bb, "ULMUMIMO", role_caps->ul_mumimo);
	blobmsg_add_u8(bb, "ULOFDMA",  role_caps->ul_ofdma);

	blobmsg_add_u32(bb, "MaxDLMUMIMO", role_caps->max_dl_mumimo);
	blobmsg_add_u32(bb, "MaxULMUMIMO", role_caps->max_ul_mumimo);
	blobmsg_add_u32(bb, "MaxDLOFDMA", role_caps->max_dl_ofdma);
	blobmsg_add_u32(bb, "MaxULOFDMA", role_caps->max_ul_ofdma);

	blobmsg_add_u8(bb, "RTS", role_caps->rts);
	blobmsg_add_u8(bb, "MURTS", role_caps->mu_rts);
	blobmsg_add_u8(bb, "MultiBSSID", role_caps->multi_bssid);
	blobmsg_add_u8(bb, "MUEDCA", role_caps->mu_edca);
	blobmsg_add_u8(bb, "TWTRequestor", role_caps->twt_requester);
	blobmsg_add_u8(bb, "TWTResponder", role_caps->twt_responder);
	blobmsg_add_u8(bb, "SpatialReuse", role_caps->spatial_reuse);
	blobmsg_add_u8(bb, "AnticipatedChannelUsage", role_caps->anticipated_ch_usage);
}


static void dump2_network_device_radio_capabilities(struct decollector_private *p,
						    struct wifi_network *net,
						    struct wifi_network_device *dev,
						    struct wifi_radio_element *radio,
						    struct blob_buf *bb)
{
	void *capabilities, *wifi6_ap_role;
	void *arr;
	int i;
	char caps_b64[32] = {0};
	const char *caps_b64_ret = NULL;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities. */
	capabilities = blobmsg_open_table(bb, "Capabilities");

	caps_b64_ret = to_base64_str((const char *)&radio->caps.ht,
				     sizeof(radio->caps.ht), caps_b64,
				     sizeof(caps_b64));
	blobmsg_add_string(bb, "HTCapabilities", caps_b64_ret ? caps_b64_ret : "");

	caps_b64_ret = to_base64_str((const char *)&radio->caps.vht,
				     sizeof(radio->caps.vht), caps_b64,
				     sizeof(caps_b64));
	blobmsg_add_string(bb, "VHTCapabilities",  caps_b64_ret ? caps_b64_ret : "");

	/* "HECapabilities" is deprecated in tr-181 2.15 and replaced by WiFi6AP/STARole */

	blobmsg_add_u32(bb, "CapableOperatingClassProfileNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "AKMFrontHaulNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "AKMBackhaulNumberOfEntries", 0); // TODO unsignedInt

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.WiFi6APRole. */
	wifi6_ap_role = blobmsg_open_table(bb, "WiFi6APRole");

	dump2_network_device_radio_capabilities_wifi6_role(&radio->caps.wifi6_ap, bb);

	blobmsg_close_table(bb, wifi6_ap_role); // WiFi6APRole

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.WiFi6bSTARole. */
	wifi6_ap_role = blobmsg_open_table(bb, "WiFi6bSTARole");

	dump2_network_device_radio_capabilities_wifi6_role(&radio->caps.wifi6_bsta, bb);

	blobmsg_close_table(bb, wifi6_ap_role); // WiFi6bSTARole

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.AKMFrontHaul.{i}. */
	arr = blobmsg_open_array(bb, "AKMFrontHaul");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "OUI", "TODO base64");
		blobmsg_add_u32(bb, "Type", 0); // TODO unsignedInt​(:255)
	}
	blobmsg_close_array(bb, arr); // AKMFrontHaul

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.AKMBackhaul.{i}. */
	arr = blobmsg_open_array(bb, "AKMBackhaul");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "OUI", "TODO base64");
		blobmsg_add_u32(bb, "Type", 0); // TODO unsignedInt​(:255)
	}
	blobmsg_close_array(bb, arr); // AKMBackhaul

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.Capabilities.CapableOperatingClassProfile.{i}. */
	arr = blobmsg_open_array(bb, "CapableOperatingClassProfile");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "Class", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "MaxTxPower", 0); // TODO int​(-127:127)
		arr = blobmsg_open_array(bb, "NonOperable");
		for (i = 0; i < 0 /* TODO */; i++) {
			 ; // list of unsignedInt​(:255)
		}
		blobmsg_close_array(bb, arr);
		blobmsg_add_u32(bb, "NumberOfNonOperChan", 0); // TODO unsignedInt
	}
	blobmsg_close_array(bb, arr); // CapableOperatingClassProfile

	blobmsg_close_table(bb, capabilities); // Capabilities
}

static void dump2_network_device_radio_opclass(struct decollector_private *p,
					       struct wifi_network *net,
					       struct wifi_network_device *dev,
					       struct wifi_radio_element *radio,
					       struct blob_buf *bb)
{
	void *arr, *arr2;
	void *t, *t2;
	int i;


	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.CurrentOperatingClassProfile.{i}. */
	arr = blobmsg_open_array(bb, "CurrentOperatingClassProfile");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_u32(bb, "Class", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "Channel", 0); // TODO unsignedInt​(:255)
		blobmsg_add_u32(bb, "TxPower", 0); // TODO int​(-127:127)
		blobmsg_add_u32(bb, "TransmitPowerLimit", 0); // TODO int​(-128:127)
		blobmsg_add_string(bb, "TimeStamp", "TODO");
	}
	blobmsg_close_array(bb, arr); // CurrentOperatingClassProfile

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.DisAllowedOpClassChannels.{i}. */
	arr = blobmsg_open_array(bb, "DisAllowedOpClassChannels");
	for (i = 0; i < 0 /* TODO */; i++) {
		blobmsg_add_string(bb, "Enable", "TODO bool");
		blobmsg_add_u32(bb, "OpClass", 0); // TODO unsignedInt​(:255)
		arr2 = blobmsg_open_array(bb, "ChannelList");
		for (i = 0; i < 0 /* TODO */; i++) {
			t2 = blobmsg_open_table(bb, "");
			 ; // list of unsignedInt​(:255)
			blobmsg_close_table(bb, t2);
		}
		blobmsg_close_array(bb, arr2);
	}
	blobmsg_close_array(bb, arr); // DisAllowedOpClassChannels

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.SpatialReuse. */
	t = blobmsg_open_table(bb, "SpatialReuse");
	blobmsg_add_u32(bb, "PartialBSSColor", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "BSSColor", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "HESIGASpatialReuseValue15Allowed", "TODO bool");
	blobmsg_add_string(bb, "SRGInformationValid", "TODO bool");
	blobmsg_add_string(bb, "NonSRGOffsetValid", "TODO bool");
	blobmsg_add_string(bb, "PSRDisallowed", "TODO bool");
	blobmsg_add_u32(bb, "NonSRGOBSSPDMaxOffset", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "SRGOBSSPDMinOffset", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "SRGOBSSPDMaxOffset", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "SRGBSSColorBitmap", "TODO hexBinary​(8:8)");
	blobmsg_add_string(bb, "SRGPartialBSSIDBitmap", "TODO hexBinary​(8:8)");
	blobmsg_add_string(bb, "NeighborBSSColorInUseBitmap", "TODO hexBinary​(8:8)");
	blobmsg_close_table(bb, t); // SpatialReuse
}



static void dump2_network_device_radio(struct decollector_private *p,
				       struct wifi_network *net,
				       struct wifi_network_device *dev,
				       struct wifi_radio_element *radio,
				       struct blob_buf *bb)
{
	struct wifi_unassoc_sta_element *usta = NULL;
	struct wifi_bss_element *bss = NULL;
	char id[32];
	void *arr, *t, *radio_table;

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}. */

	radio_table = blobmsg_open_table(bb, "");

	blobmsg_add_string(bb, "ID",
			   to_base64_str((const char *)radio->macaddr, sizeof(radio->macaddr), id, sizeof(id)));
	blobmsg_add_string(bb, "Enabled", radio->enabled ? "true" : "false");
	blobmsg_add_u32(bb, "Noise", radio->anpi);
	blobmsg_add_u32(bb, "Utilization", radio->total_utilization);
	blobmsg_add_u32(bb, "Transmit", radio->tx_utilization);
	blobmsg_add_u32(bb, "ReceiveSelf", radio->rx_utilization);
	blobmsg_add_u32(bb, "ReceiveOther", radio->other_utilization);
	blobmsg_add_u8(bb, "TrafficSeparationCombinedFronthaul", radio->ts_combined_fronthaul);
	blobmsg_add_u8(bb, "TrafficSeparationCombinedBackhaul", radio->ts_combined_backhaul);

	// TODO: Multi-AP Policy Config Request / Steering Policy TLV
	blobmsg_add_u32(bb, "SteeringPolicy", 0); // TODO unsignedInt​(:2)
	blobmsg_add_u32(bb, "ChannelUtilizationThreshold", 0); // TODO unsignedInt​(:255)
	blobmsg_add_u32(bb, "RCPISteeringThreshold", 0); // TODO unsignedInt​(:220)

	// TODO: Multi-AP Policy Config Request / Metric Reporting Policy TLV
	blobmsg_add_u32(bb, "STAReportingRCPIThreshold", 0); // TODO unsignedInt​(:220)
	blobmsg_add_u32(bb, "STAReportingRCPIHysteresisMarginOverride", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "ChannelUtilizationReportingThreshold", 0); // TODO unsignedInt
	blobmsg_add_string(bb, "AssociatedSTATrafficStatsInclusionPolicy", "TODO bool");
	blobmsg_add_string(bb, "AssociatedSTALinkMetricsInclusionPolicy", "TODO bool");

	blobmsg_add_string(bb, "ChipsetVendor", radio->vendor);

	// TODO: Multi-AP Policy Config Request / Metric Reporting Policy TLV
	blobmsg_add_string(bb, "APMetricsWiFi6", "TODO bool");

	blobmsg_add_u32(bb, "CurrentOperatingClassProfileNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "UnassociatedSTANumberOfEntries", radio->num_unassoc_sta);
	blobmsg_add_u32(bb, "BSSNumberOfEntries", radio->num_bss);
	blobmsg_add_u32(bb, "ScanResultNumberOfEntries", radio->num_scanresult);
	blobmsg_add_u32(bb, "DisAllowedOpClassChannelsNumberOfEntries", 0); // TODO unsignedInt

	dump2_network_device_radio_scanresult(p, net, dev, radio, bb);
	dump2_network_device_radio_backhaulsta(p, net, dev, radio, bb);
	dump2_network_device_radio_scancapability(p, net, dev, radio, bb);
	dump2_network_device_radio_caccapability(p, net, dev, radio, bb);
	dump2_network_device_radio_capabilities(p, net, dev, radio, bb);
	dump2_network_device_radio_opclass(p, net, dev, radio, bb);

	arr = blobmsg_open_array(bb, "BSSList");
	list_for_each_entry(bss, &radio->bsslist, list) {
		dump2_network_device_radio_bss(p, net, dev, radio, bss, bb);
	}
	blobmsg_close_array(bb, arr);

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.UnassociatedSTA.{i}. */
	arr = blobmsg_open_array(bb, "UnassociatedSTA");
	list_for_each_entry(usta, &radio->unassoc_stalist, list) {
		char usta_macaddr[18] = { 0 };

		hwaddr_ntoa(usta->macaddr, usta_macaddr);
		blobmsg_add_string(bb, "MACAddress", usta_macaddr);
		blobmsg_add_u32(bb, "SignalStrength", usta->rcpi);
	}
	blobmsg_close_array(bb, arr); // UnassociatedSTA

	/* Device.WiFi.DataElements.Network.Device.{i}.Radio.{i}.MultiAPRadio. */
	t = blobmsg_open_table(bb, "MultiAPRadio");
	//Unknown type  list​(1024) of unsignedInt TOKEN RadarDetections
	blobmsg_close_table(bb, t); // MultiAPRadio

	blobmsg_close_table(bb, radio_table);
}

static void dump2_network_device(struct decollector_private *p, struct wifi_network *net,
				 struct wifi_network_device *dev, struct blob_buf *bb)
{
	struct wifi_radio_element *radio = NULL;
	struct mac_address_element *macaddr = NULL;
	char macstr[18] = { 0 };
	char ap_caps_encoded[8] = { 0 };
	void *arr, *device_table;

	/* Device.WiFi.DataElements.Network.Device.{i}. */
	device_table = blobmsg_open_table(bb, "");

	hwaddr_ntoa(dev->macaddr, macstr);
	blobmsg_add_string(bb, "ID", macstr);

	blobmsg_add_string(bb, "MultiAPCapabilities",
			   to_base64_str((const char *)&dev->multiap_caps, sizeof(dev->multiap_caps),
					 ap_caps_encoded, sizeof(ap_caps_encoded)));

	blobmsg_add_u32(bb, "CollectionInterval", dev->collect_int);

	blobmsg_add_u8(bb, "ReportUnsuccessfulAssociations", dev->report.sta_assocfails);
	blobmsg_add_u32(bb, "MaxReportingRate", dev->report.sta_assocfails_rate);
	blobmsg_add_u32(bb, "APMetricsReportingInterval", dev->report.ap_metrics_int);

	blobmsg_add_string(bb, "Manufacturer", "IOPSYS");
	blobmsg_add_string(bb, "SerialNumber", dev->serial);
	blobmsg_add_string(bb, "ManufacturerModel", "TODO"); // TODO: string
	blobmsg_add_string(bb, "SoftwareVersion", dev->swversion);
	blobmsg_add_string(bb, "ExecutionEnv", dev->execenv);

	// TODO: Service Prioritization Request / DSCP Mapping Table TLV
	blobmsg_add_string(bb, "DSCPMap", "TODO"); //TODO: hexBinary​(64) string

	blobmsg_add_u32(bb, "MaxPrioritizationRules", dev->max_prules);
	blobmsg_add_u8(bb, "PrioritizationSupport", dev->support_sp);
	blobmsg_add_u32(bb, "MaxVIDs", dev->max_vids);

	// TODO:  Multi-AP Policy Config Request / Metric Reporting Policy TLV
	blobmsg_add_string(bb, "APMetricsWiFi6", "TODO bool"); // TODO boolean
	blobmsg_add_string(bb, "CountryCode", dev->country_code);

	arr = blobmsg_open_array(bb, "LocalSteeringDisallowedSTAList");
	list_for_each_entry(macaddr, &dev->sta_steer_disallowlist, list) {
		hwaddr_ntoa(macaddr->macaddr, macstr);
		blobmsg_add_string(bb, "", macstr);
	}
	blobmsg_close_array(bb, arr);

	arr = blobmsg_open_array(bb, "BTMSteeringDisalloywedSTAList");
	list_for_each_entry(macaddr, &dev->sta_btmsteer_disallowlist, list) {
		hwaddr_ntoa(macaddr->macaddr, macstr);
		blobmsg_add_string(bb, "", macstr);
	}
	blobmsg_close_array(bb, arr);

	blobmsg_add_u8(bb, "DFSEnable", dev->dfs_enabled);

	blobmsg_add_u8(bb, "ReportIndependentScans",  dev->report.independent_scans);

	blobmsg_add_u8(bb, "STASteeringState", dev->sta_steering_state);
	blobmsg_add_string(bb, "CoordinatedCACAllowed", "TODO bool"); // TODO boolean
	blobmsg_add_string(bb, "TrafficSeparationAllowed", "TODO bool"); // TODO boolean
	blobmsg_add_string(bb, "ServicePrioritizationAllowed", "TODO bool"); // TODO boolean

	blobmsg_add_u32(bb, "RadioNumberOfEntries", dev->num_radios);

	// TODO: see dump2_network_device_default8021q
	blobmsg_add_u32(bb, "Default8021QNumberOfEntries",  net->primary_vid ? 1 : 0); // TODO unsignedInt

	// TODO: see dump2_network_device_ssidtovidmapping
	blobmsg_add_u32(bb, "SSIDtoVIDMappingNumberOfEntries", list_count(&net->ssidlist));

	blobmsg_add_u32(bb, "CACStatusNumberOfEntries", list_count(&dev->cac_statuslist));

	blobmsg_add_u32(bb, "IEEE1905SecurityNumberOfEntries", dev->i1905_seccap.caps_valid ? 1 : 0);
	blobmsg_add_u32(bb, "SPRuleNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "AnticipatedChannelsNumberOfEntries", 0); // TODO unsignedInt
	blobmsg_add_u32(bb, "AnticipatedChannelUsageNumberOfEntries", 0); // TODO unsignedInt

	dump2_network_device_default8021q(p, net, dev, bb);
	dump2_network_device_ssidtovidmapping(p, net, dev, bb);
	dump2_network_device_cacstatus(p, net, dev, bb);
	dump2_network_device_sprule(p, net, dev, bb);
	dump2_network_device_ieee1905security(p, net, dev, bb);
	dump2_network_device_anticipatedchannels(p, net, dev, bb);
	dump2_network_device_anticipatedchannelsusage(p, net, dev, bb);
	dump2_network_device_multiapdevice(p, net, dev, bb);

	arr = blobmsg_open_array(bb, "RadioList");
	list_for_each_entry(radio, &dev->radiolist, list) {
		dump2_network_device_radio(p, net, dev, radio, bb);
	}
	blobmsg_close_array(bb, arr);

	blobmsg_close_table(bb, device_table);
}

static void dump2_network_ssid(struct blob_buf *bb, struct wifi_network_ssid *ssid)
{
	void *t;
	char band_list[32] = { 0 };

	/* Device.WiFi.DataElements.Network.SSID.{i}. */
	t = blobmsg_open_table(bb, "");

	blobmsg_add_string(bb, "SSID", (char *)ssid->ssid); // TODO: safe ssid - non printable chars
	blobmsg_add_string(bb, "Band",
			   create_band_list(band_list, sizeof(band_list), ssid->band));

	blobmsg_close_table(bb, t);
}

static void dump2_network_multi_ap_steering_summary_stats(
	struct blob_buf *bb, const struct wifi_steer_summary *steer_stats)
{
	void *t;
	/* Device.WiFi.DataElements.Network.MultiAPSteeringSummaryStats. */

	t = blobmsg_open_table(bb, "MultiAPSteeringSummaryStats");

	blobmsg_add_u64(bb, "NoCandidateAPFailures", steer_stats->no_candidate_cnt);
	blobmsg_add_u64(bb, "BlacklistAttempts", steer_stats->blacklist_attempt_cnt);
	blobmsg_add_u64(bb, "BlacklistSuccesses", steer_stats->blacklist_success_cnt);
	blobmsg_add_u64(bb, "BlacklistFailures", steer_stats->blacklist_failure_cnt);
	blobmsg_add_u64(bb, "BTMAttempts", steer_stats->btm_attempt_cnt);
	blobmsg_add_u64(bb, "BTMSuccesses", steer_stats->btm_success_cnt);
	blobmsg_add_u64(bb, "BTMFailures", steer_stats->btm_failure_cnt);
	blobmsg_add_u64(bb, "BTMQueryResponses", steer_stats->btm_query_resp_cnt);

	blobmsg_close_table(bb, t); // MultiAPSteeringSummaryStats
}

static void dump2_network(struct decollector_private *p, struct wifi_network *net,
			  struct blob_buf *bb)
{
	struct wifi_network_device *dev = NULL;
	struct wifi_network_ssid *ssid = NULL;
	char cntlr_id[18] = { 0 };
	char timestamp[32] = { 0 };
	void *ssid_arr, *device_arr;
	uint32_t available_nodes;

	/* Device.WiFi.DataElements.Network. */

	blobmsg_add_string(bb, "ID", net->id);
	blobmsg_add_string(bb, "TimeStamp", get_timestamp(NULL, timestamp));
	if (!p->cfg.is_sap_mode) {
		hwaddr_ntoa(net->cntlr_id, cntlr_id);
		blobmsg_add_string(bb, "ControllerID", cntlr_id);
	}

	available_nodes = net->num_devices;
	list_for_each_entry(dev, &net->devicelist, list) {
		if (dev->invalid)
			available_nodes--;
	}
	blobmsg_add_u32(bb, "DeviceNumberOfEntries", available_nodes);

	// TODO:   Multi-AP Policy Config Request / QoS Management Policy TLV
	blobmsg_add_string(bb, "MSCSDisallowedStaList", "" /* TODO: comma_separated_mac_list */);
	blobmsg_add_string(bb, "SCSDisallowedStaList", "" /* TODO: comma_separated_mac_list */);

	blobmsg_add_u32(bb, "SSIDNumberOfEntries", net->num_ssid);

	ssid_arr = blobmsg_open_array(bb, "SSID");
	list_for_each_entry(ssid, &net->ssidlist, list) {
		dump2_network_ssid(bb, ssid);
	}
	blobmsg_close_array(bb, ssid_arr);

	dump2_network_multi_ap_steering_summary_stats(bb, &net->steer_summary);

	device_arr = blobmsg_open_array(bb, "DeviceList");
	list_for_each_entry(dev, &net->devicelist, list) {
		if (dev->invalid)
			continue;
		dump2_network_device(p, net, dev, bb);
	}
	blobmsg_close_array(bb, device_arr);
}

int decollector_dump2_all(struct decollector_private *p, struct blob_buf *bb)
{
	struct wifi_data_element *dm = p->dm;
	char dstr[32] = {0};
	char tsp[32] = {0};
	void *t, *t2;
	void *arr;

	blobmsg_add_string(bb, "date", get_date(NULL, dstr));
	blobmsg_add_string(bb, "version", WIFI_DATAELEMENTS_VER_2_0);
	blobmsg_add_string(bb, "description", "WFA Data Elements");
	blobmsg_add_string(bb, "TimeStamp", get_timestamp(NULL, tsp));
	blobmsg_add_string(bb, "name", "wifi");

	arr = blobmsg_open_array(bb, "data");
	t = blobmsg_open_table(bb, "");
	t2 = blobmsg_open_table(bb, "wfa-dataelements:Network");

	dump2_network(p, &dm->network, bb);

	blobmsg_close_table(bb, t2);
	blobmsg_close_table(bb, t);
	blobmsg_close_array(bb, arr);

	return 0;
}
